#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from datetime import datetime

# import interfaces
from pyams_security.interfaces import ISecurityManager
from pyams_workflow.interfaces import IWorkflowManagedContent, IWorkflowPublicationInfo, IWorkflow, IWorkflowVersions, \
    IWorkflowPublicationSupport, IObjectClonedEvent, IWorkflowState
from zope.annotation.interfaces import IAnnotations

# import packages
from persistent import Persistent
from pyams_utils.adapter import adapter_config
from pyams_utils.date import format_datetime
from pyams_utils.registry import get_utility, query_utility
from pyams_utils.request import check_request
from pyams_utils.timezone import gmtime, tztime
from pyams_utils.traversing import get_parent
from pyams_workflow.versions import WorkflowHistoryItem
from pyramid.events import subscriber
from pyramid.threadlocal import get_current_registry
from zope.container.contained import Contained
from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty

from pyams_workflow import _


@implementer(IWorkflowPublicationInfo)
class WorkflowContentPublicationInfo(Persistent, Contained):
    """Workflow content info"""

    _publication_date = FieldProperty(IWorkflowPublicationInfo['publication_date'])
    publisher = FieldProperty(IWorkflowPublicationInfo['publisher'])
    _first_publication_date = FieldProperty(IWorkflowPublicationInfo['first_publication_date'])
    _publication_effective_date = FieldProperty(IWorkflowPublicationInfo['publication_effective_date'])
    _publication_expiration_date = FieldProperty(IWorkflowPublicationInfo['publication_expiration_date'])

    @property
    def publication_date(self):
        return self._publication_date

    @publication_date.setter
    def publication_date(self, value):
        self._publication_date = gmtime(value)

    @property
    def publication(self):
        request = check_request()
        translate = request.localizer.translate
        auth = get_utility(ISecurityManager)
        return translate(_('{date} by {principal}')).format(
            date=format_datetime(tztime(IWorkflowPublicationInfo(self).publication_date), request=request),
            principal=translate(auth.get_principal(self.publisher).title))

    @property
    def first_publication_date(self):
        return self._first_publication_date

    @property
    def publication_effective_date(self):
        return self._publication_effective_date

    @publication_effective_date.setter
    def publication_effective_date(self, value):
        self._publication_effective_date = gmtime(value)
        if value and ((self._first_publication_date is None) or
                      (self._first_publication_date > self._publication_effective_date)):
            self._first_publication_date = self._publication_effective_date

    @property
    def publication_expiration_date(self):
        return self._publication_expiration_date

    @publication_expiration_date.setter
    def publication_expiration_date(self, value):
        self._publication_expiration_date = gmtime(value)

    def reset(self):
        self._publication_date = None
        self._first_publication_date = None
        self._publication_effective_date = None
        self._publication_expiration_date = None

    def is_published(self):
        # check is parent is published
        parent = get_parent(self.__parent__, IWorkflowPublicationSupport, allow_context=False)
        if (parent is not None) and not IWorkflowPublicationInfo(parent).is_published():
            return False
        # associated workflow?
        wf_name = IWorkflowManagedContent(self.__parent__).workflow_name
        if not wf_name:
            return True
        # published state(s) in workflow?
        workflow = query_utility(IWorkflow, name=wf_name)
        if (workflow is None) or not workflow.published_states:
            return False
        # check content versions
        versions = IWorkflowVersions(self.__parent__)
        if not (versions or versions.get_versions(workflow.published_states)):
            return False
        now = datetime.utcnow()
        return (self.publication_effective_date is not None) and \
               (self.publication_effective_date <= now) and \
               ((self.publication_expiration_date is None) or
                (self.publication_expiration_date >= now))

    def is_visible(self, request=None):
        # associated workflow?
        content = IWorkflowManagedContent(self.__parent__)
        wf_name = content.workflow_name
        if not wf_name:
            return True
        # check workflow?
        if content.view_permission:
            if request is None:
                request = check_request()
            if not request.has_permission(content.view_permission, context=self.__parent__):
                return False
        return self.is_published()


WORKFLOW_CONTENT_KEY = 'pyams_workflow.content_info'


@adapter_config(context=IWorkflowPublicationSupport, provides=IWorkflowPublicationInfo)
def WorkflowContentPublicationInfoFactory(context):
    """Workflow content info factory"""
    annotations = IAnnotations(context)
    info = annotations.get(WORKFLOW_CONTENT_KEY)
    if info is None:
        info = annotations[WORKFLOW_CONTENT_KEY] = WorkflowContentPublicationInfo()
        registry = get_current_registry()
        registry.notify(ObjectCreatedEvent(info))
        locate(info, context)
    return info


@subscriber(IObjectClonedEvent)
def handle_cloned_object(event):
    """Add comment when an object is cloned"""
    request = check_request()
    translate = request.localizer.translate
    source_state = IWorkflowState(event.source)
    target_state = IWorkflowState(event.object)
    item = WorkflowHistoryItem(date=datetime.utcnow(),
                               source_state=source_state.state,
                               principal=request.principal.id,
                               comment=translate(_("Clone created from version {source}")).format(
                                   source=source_state.version_id))
    target_state.history.append(item)


@subscriber(IObjectClonedEvent, context_selector=IWorkflowPublicationSupport)
def handle_cloned_publication_support(event):
    """Reset publication info when an object is cloned"""
    IWorkflowPublicationInfo(event.object).reset()
